from math import sqrt, pi, sin, asin, radians, cos, acos
from random import triangular, randint, choice, random, randrange, uniform
from collections import deque
from glob import glob
from os.path import join, basename

from pygame import Surface, PixelArray, Color, Rect
from pygame.font import Font
from pygame.mixer import Sound
from pygame.draw import line, polygon, circle
from pygame.time import get_ticks
from pygame.image import load
from pygame.event import clear
from pygame.transform import rotate
from pygame.locals import *

class SoundEffect(Sound):

    def __init__(self, path, volume):
        Sound.__init__(self, path)
        self.set_volume(volume)

from seam.pgfw.Game import Game
from seam.pgfw.GameChild import GameChild
from seam.pgfw.Sprite import Sprite
from seam.pgfw.Animation import Animation

class Seam(Game, Animation):

    WIRE_SPEED = 2
    ENEMY_SPAWN_RATE = .0008
    INITIAL_POPULATION = 5
    ENEMY_SPEED = 4
    ENEMY_MOVE_PROBABILITY = .01
    POWER_UP_SUPPLY_RATE = .005
    TRANSPARENT_COLOR = Color(255, 0, 255)

    def __init__(self):
        Game.__init__(self)
        Animation.__init__(self, self)
        self.delegate = self.get_game().delegate
        self.capturing = False
        self.ground = ground = Surface(self.get_screen().get_size())
        ground.fill((225, 192, 87))
        ground.set_colorkey(self.TRANSPARENT_COLOR)
        self.map = Map(self)
        self.timer = Timer(self)
        self.wires = Wires(self)
        self.ships = Ships(self)
        self.net = Net(self)
        self.enemies = Enemies(self)
        self.life_meter = LifeMeter(self)
        self.grid = Grid(self)
        # self.power_ups = PowerUps(self)
        self.success_message = EndMessage(self, "SUCCESS", 0)
        self.failure_message = EndMessage(self, "GAME OVER", 0)
        self.instructions = Instructions(self)
        self.high_scores = HighScores(self)
        self.next_sound_effect = SoundEffect(self.get_resource("next.wav"), .9)
        self.introduction_audio = Sound(self.get_resource("Beam.ogg"))
        self.game_audio = Sound(self.get_resource("Mouth.ogg"))
        self.end_audio = Sound(self.get_resource("Tode.ogg"))
        self.register(self.release_game_over_wait_period,
                      self.release_instructions_display_wait_period)
        self.reset()
        self.subscribe(self.respond)
        clear()

    def reset(self):
        self.game_over = False
        self.success = False
        self.game_over_wait_period_over = True
        self.displaying_instructions = True
        self.instructions_display_wait_period_over = False
        self.play(self.release_instructions_display_wait_period, delay=800,
                  play_once=True)
        self.map.reset()
        self.end_audio.fadeout(1000)
        self.game_audio.fadeout(1000)
        self.introduction_audio.play(-1, 0, 1000)
        self.timer.reset()
        self.ships.reset()
        self.wires.reset()
        self.net.reset()
        self.enemies.reset()
        self.life_meter.reset()
        self.grid.reset()
        self.success_message.reset()
        self.failure_message.reset()
        self.instructions.reset()
        self.high_scores.reset()
        # self.power_ups.reset()

    def release_instructions_display_wait_period(self):
        self.instructions_display_wait_period_over = True

    def respond(self, event):
        compare = self.delegate.compare
        if not self.game_over and not self.displaying_instructions:
            if compare(event, "act"):
                self.net.start()
            elif compare(event, "act", True) and self.net.active:
                self.net.capture()
        if compare(event, "reset-game"):
            self.reset()
        elif self.game_over and self.game_over_wait_period_over and \
                 compare(event, "any"):
            self.reset()
            self.next_sound_effect.play()
            self.end_audio.fadeout(1000)
        elif self.displaying_instructions and \
             self.instructions_display_wait_period_over and \
             compare(event, "any"):
            self.displaying_instructions = False
            self.instructions.deactivate()
            self.high_scores.deactivate()
            self.timer.start()
            self.enemies.spawn_initial()
            self.enemies.activate_spawning()
            # self.power_ups.activate()
            self.next_sound_effect.play()
            self.introduction_audio.fadeout(1000)
            self.game_audio.play(-1, 0, 0)

    def end_game(self, success):
        self.enemies.halt_spawning()
        self.ships.cancel_motion()
        self.net.deactivate()
        # self.power_ups.deactivate()
        self.game_audio.fadeout(1000)
        self.end_audio.play(-1, 0, 1000)
        self.timer.running = False
        self.game_over = True
        self.game_over_wait_period_over = False
        self.play(self.release_game_over_wait_period, delay=1200,
                  play_once=True)
        if success:
            scores = open(self.get_resource("scores"), "a")
            string = str(self.grid.ratio)
            scores.write(string + "\n")
            scores.close()
            self.success_message.activate()
            self.high_scores.refresh(string)
            self.next_sound_effect.play()
        else:
            self.failure_message.activate()
            self.high_scores.cancel_blink()
            self.enemies.hit_sound_effect.play()
        self.success = success

    def release_game_over_wait_period(self):
        self.game_over_wait_period_over = True

    def update(self):
        self.map.update()
        self.draw_ground()
        self.wires.update()
        self.ships.shift()
        self.net.update()
        self.ships.update()
        self.enemies.update()
        # self.power_ups.update()
        if not self.game_over and self.life_meter.lives == 0:
            self.end_game(False)
        self.life_meter.update()
        self.timer.update()
        self.grid.update()
        self.failure_message.update()
        self.success_message.update()
        self.instructions.update()
        self.high_scores.update()
        Animation.update(self)

    def draw_ground(self):
        opened = self.ground.copy()
        for rect in self.map.revealed:
            opened.fill(self.TRANSPARENT_COLOR, rect)
        self.get_screen().blit(opened, (0, 0))


class Map(Sprite):

    SHADOW_THICKNESS = 8
    SHADOW_ALPHA = 178
    HORIZONTAL, VERTICAL = range(2)

    def __init__(self, parent):
        Sprite.__init__(self, parent, 125)
        self.set_frames()
        self.set_shadow_strips()

    def set_frames(self):
        for path in sorted(glob(join(self.get_resource("display",
                                                       "map-frame-path"),
                                     "*.png"))):
            self.add_frame(load(path).convert())

    def set_shadow_strips(self):
        tile = Surface([self.SHADOW_THICKNESS] * 2, SRCALPHA)
        tw, th = tile.get_size()
        for y in xrange(th):
            tile.fill((0, 0, 0, self.SHADOW_ALPHA * (1 - float(y) / (th - 1))),
                      (0, y, tw, 1))
        ds = self.get_display_surface()
        horizontal, vertical = self.shadow_strips = (Surface((ds.get_width(),
                                                              th), SRCALPHA),
                                                     Surface((tw,
                                                              ds.get_height()),
                                                             SRCALPHA))
        for x in xrange(0, horizontal.get_width(), tw):
            horizontal.blit(tile, (x, 0))
        for y in xrange(0, vertical.get_height(), tw):
            vertical.blit(rotate(tile, 90), (0, y))
        diagonal = self.diagonal = tile.copy()
        for w in xrange(tw, 0, -1):
            diagonal.fill((0, 0, 0, self.SHADOW_ALPHA * (1 - float(w) / tw)),
                          (0, 0, w, w))
        # we had a good dog sperm
        # see you in the afterlife
        # psychotic thoughts of blowing people away one by one
        # BLAT BLAT BLAT
        # everybody goes down
        # bananas I'm losing it
        # I'm going bananas
        # there's nothing but a hole
        # and 12 packs of crayons
        # snuffed out at 187 years old

    def reset(self):
        self.revealed = []
        self.shadow_markers = []

    def reveal(self, rect):
        self.revealed.append(rect)
        self.set_shadow_markers()

    def set_shadow_markers(self):
        shadow_markers = self.shadow_markers = [], []
        for rect in self.revealed:
            for orientation in self.HORIZONTAL, self.VERTICAL:
                if orientation == self.HORIZONTAL:
                    keep = (rect.left, rect.right),
                else:
                    keep = (rect.top, rect.bottom),
                unset = []
                for other in self.revealed:
                    if rect.colliderect(other) or rect.right == other.left:
                        if orientation == self.HORIZONTAL:
                            if other.top < rect.top < other.bottom:
                                unset.append((other.left, other.right))
                        else:
                            if other.left < rect.left < other.right:
                                unset.append((other.top, other.bottom))
                for other in unset:
                    r_keep = []
                    for low, high in keep:
                        if other[0] > high or other[1] < low:
                            r_keep.append((low, high))
                        else:
                            under = low, other[0]
                            over = other[1], high
                            if under[0] < under[1]:
                                r_keep.append(under)
                            if over[0] < over[1]:
                                r_keep.append(over)
                    keep = r_keep
                shadow_markers[orientation].append((rect, keep))

    def update(self):
        Sprite.update(self, self.revealed)
        ds = self.display_surface
        for ii, markers in enumerate(self.shadow_markers):
            strip = self.shadow_strips[ii]
            for rect, keep in markers:
                for low, high in keep:
                    if ii == self.HORIZONTAL:
                        position = low, rect.top
                        w, h = high - low, strip.get_height()
                    else:
                        position = rect.left, low
                        w, h = strip.get_width(), high - low
                    ds.blit(strip, position, (0, 0, w, h))
                    # ds.blit(strip, position)
                    if ii == self.HORIZONTAL:
                        ds.blit(self.diagonal, (high, rect.top))


class Timer(GameChild):

    limit = 99999

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        self.display_surface = self.get_display_surface()
        self.font = Font(self.get_resource("display", "timer-font"), 34)

    def reset(self):
        self.elapsed = 0
        self.running = False
        self.last_ticks = None

    def start(self):
        self.running = True
        self.last_ticks = None

    def update(self):
        ticks = get_ticks()
        if self.running and self.last_ticks is not None:
            self.elapsed += ticks - self.last_ticks
            if self.elapsed >= self.limit:
                self.elapsed = self.limit
                self.parent.end_game(True)
        self.last_ticks = ticks
        remaining = str(int(self.limit - self.elapsed) / 1000)
        for ii, digit in enumerate(remaining):
            center = self.display_surface.get_rect().centerx
            w, h = 31, 40
            background = Surface((w, h))
            background_color = (255, 255, 255)
            background.fill(background_color)
            background.fill((128, 90, 35), (0, 0, w, 2))
            rect = background.get_rect()
            rect.centerx = center + (ii - ((len(remaining) - 1) / 2.0)) * w
            glyph = self.font.render(digit, True, (31, 31, 31))
            gr = glyph.get_rect()
            gr.center = rect.w / 2, rect.h / 2 + 1
            background.blit(glyph, gr)
            self.display_surface.blit(background, rect)


class Wires(GameChild, list):

    AMPLITUDES = 8, 10, 13, 16, 18, 16, 13, 10
    WIDTH = 9592
    LEFT, RIGHT = -1, 1

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        list.__init__(self, (Wire(self, self.LEFT), Wire(self, self.RIGHT)))

    def reset(self):
        for wire in self:
            wire.reset()

    def update(self):
        for wire in self:
            wire.update()


class Wire(Sprite):

    TRANSPARENT_COLOR = 255, 0, 255

    def __init__(self, parent, heading):
        Sprite.__init__(self, parent)
        self.heading = heading
        self.display_surface = self.get_display_surface()
        self.color = (255, 255, 255) if heading == parent.LEFT else \
                     (108, 108, 108)
        self.reset()
        self.add_location(offset=(self.location.w, 0))
        self.register(self.increment_amplitude_index, interval=100)
        self.play(self.increment_amplitude_index)

    def reset(self):
        self.amplitude_index = 0
        # self.clear_frames()
        self.set_points()
        # self.set_frames()

    def set_points(self):
        points = self.points = []
        next_jump = 0
        x = randint(-90, 90)
        while x < self.parent.WIDTH:
            if next_jump == 0 and x < self.parent.WIDTH - 800:
                random_point = self.get_random_point(x)
                while len(points) > 0 and \
                          abs(random_point[1] - points[-1][1]) < 100:
                    random_point = self.get_random_point(x)
                points.append(random_point)
                next_jump = int(round(triangular(2, 5)))
            x += randint(218, 402)
            points.append((x, self.get_random_y(points[-1][1])))
            next_jump -= 1

    def get_random_point(self, x):
        return x, self.get_random_y()

    def get_random_y(self, previous=None):
        if previous is None:
            previous = self.display_surface.get_rect().centery
        deviation = 73
        y = previous + abs(deviation - triangular(-deviation, deviation)) * \
            choice((1, -1))
        upper, lower = 12, self.display_surface.get_height() - 12
        if y < upper:
            y += (upper - y) * 2
        elif y > lower:
            y -= (y - lower) * 2
        return y

    def set_frames(self):
        for amplitude in self.parent.AMPLITUDES:
            frame = Surface((self.parent.WIDTH,
                             self.display_surface.get_height()))
            frame.fill(self.TRANSPARENT_COLOR)
            frame.set_colorkey(self.TRANSPARENT_COLOR)
            step = 10
            px, py = 0, self.get_y(0, amplitude)
            for x in xrange(step, self.parent.WIDTH, step):
                y = self.get_y(x, amplitude)
                if abs(y - py) < 20:
                    line(frame, self.color, (px, py), (x, y))
                px, py = x, y
            self.add_frame(frame)

    def get_y(self, x, amplitude=None):
        if amplitude is None:
            amplitude = self.get_current_amplitude()
        while x < 0:
            x += self.parent.WIDTH
        x %= self.parent.WIDTH
        index = 0
        points = self.points
        while x > points[index + 1][0]:
            index += 1
        px, py = points[index]
        nx, ny = points[index + 1]
        dx = nx - px
        distance = sqrt(dx ** 2 + (ny - py) ** 2)
        h = int(round(distance * float(x - px) / dx))
        cycles = 4 * pi / distance
        beta = radians(90) - asin((nx - px) / distance)
        if ny < py:
            beta *= -1
        sx = amplitude * sin(cycles * h) + py
        return sin(beta) * h - cos(beta) * (sx - py) + py

    def get_current_amplitude(self):
        return self.parent.AMPLITUDES[self.amplitude_index]

    def increment_amplitude_index(self):
        self.amplitude_index += 1
        if self.amplitude_index >= len(self.parent.AMPLITUDES):
            self.amplitude_index = 0

    def update(self):
        self.move(self.heading * self.parent.parent.WIRE_SPEED)
        if self.location.right < 0:
            self.move(self.location.width)
        if self.location.left > 0:
            self.move(-self.location.width)
        step = 10
        offset = -self.location.left
        px, py = offset, self.get_y(offset)
        ds = self.display_surface
        for x in xrange(offset, offset + ds.get_width() + step, step):
            y = self.get_y(x)
            if abs(y - py) < 20:
                line(ds, self.color, (px - offset, py), (x - offset, y))
            px, py = x, y
        Sprite.update(self)


class Ships(GameChild, list):

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        list.__init__(self, [Ship(self, Ship.OUTER), Ship(self, Ship.INNER)])

    def reset(self):
        for ship in self:
            ship.reset()

    def shift(self):
        for ship in self:
            ship.shift()

    def cancel_motion(self):
        for ship in self:
            ship.cancel_motion()

    def update(self):
        for ship in self:
            ship.update()


class Ship(Sprite):

    OUTER, INNER = range(2)
    LEFT, RIGHT = range(2)

    def __init__(self, parent, position):
        Sprite.__init__(self, parent, 240)
        self.position = position
        self.wire = parent.parent.wires[0] if \
                    position == self.OUTER else parent.parent.wires[1]
        self.display_surface = self.get_display_surface()
        self.delegate = self.get_delegate()
        self.set_frames()
        self.subscribe(self.respond)
        self.cancel_motion()

    def cancel_motion(self):
        self.moving = [False, False]

    def reset(self):
        self.remove_locations()
        self.location.centerx = self.display_surface.get_rect().centerx
        self.add_location(offset=(self.display_surface.get_width(), 0))

    def set_frames(self):
        max_width = 18
        hue = 200 if self.position == self.INNER else 320
        foreground = Color(0, 0, 0)
        background = Color(0, 0, 0)
        for ii, width in enumerate(xrange(max_width, 10, -2)):
            frame = Surface((max_width, max_width))
            frame.set_colorkey((0, 0, 0))
            fr = frame.get_rect()
            rect = Rect((0, 0, width - 2, width - 2))
            rect.center = fr.center
            foreground.hsla = hue, 100, 85, 100
            background.hsla = hue, 100, 30, 100
            polygon(frame, background, ((rect.left, rect.top),
                                        (rect.right, rect.top),
                                        (rect.right, rect.bottom),
                                        (rect.left, rect.bottom)), 1)
            rect.move_ip(-1, -1)
            polygon(frame, foreground, ((rect.left, rect.top),
                                        (rect.right, rect.top),
                                        (rect.right, rect.bottom),
                                        (rect.left, rect.bottom)), 1)
            reticule = (255, 0, 0) if self.position == self.INNER else \
                       (255, 255, 0)
            if ii % 2 == 1:
                polygon(frame, reticule, ((fr.centerx - 2, fr.centery - 2),
                                          (fr.centerx + 1, fr.centery - 2),
                                          (fr.centerx + 1, fr.centery + 1),
                                          (fr.centerx - 2, fr.centery + 1)))
            hue = (hue - 16) % 360
            self.add_frame(frame)

    def respond(self, event):
        if not self.parent.parent.game_over and not \
               self.parent.parent.displaying_instructions:
            compare = self.delegate.compare
            if compare(event, "right") or compare(event, "right", True):
                self.moving[self.RIGHT] = not event.cancel
            elif compare(event, "left") or compare(event, "left", True):
                self.moving[self.LEFT] = not event.cancel

    def shift(self):
        speed = 6
        if self.moving[self.LEFT]:
            self.move(-speed)
        if self.moving[self.RIGHT]:
            self.move(speed)
        bound = self.display_surface.get_width()
        if self.location.right < 0:
            self.move(bound)
        if self.location.left > bound:
            self.move(-bound)
        for location in self.locations:
            location.centery = self.wire.get_y(self.location.centerx - \
                                               self.wire.location.left)


class Net(GameChild):

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        ds = self.display_surface = self.get_display_surface()
        self.hues = deque(xrange(0, 360, 40))
        self.growing_sound_effect = SoundEffect(self.get_resource("grow.wav"),
                                                1)
        self.capture_sound_effect = SoundEffect(self.\
                                    get_resource("capture.wav"), 1)

    def reset(self):
        self.deactivate()
        self.init_rect()

    def init_rect(self):
        rect = self.rect = Rect(0, 0, 0, 0)
        return rect

    def deactivate(self):
        self.active = False
        self.growing_sound_effect.stop()

    def start(self):
        self.ox, self.oy = 19, 19
        self.activate()

    def activate(self):
        self.active = True
        self.growing_sound_effect.play(-1)

    def capture(self):
        self.deactivate()
        rect = self.rect.clip(self.display_surface.get_rect())
        self.parent.map.reveal(self.rect)
        self.parent.grid.add(rect)
        self.capture_sound_effect.play()

    def update(self):
        if self.active:
            self.ox += .375
            self.oy += .375
            rect = self.init_rect()
            ships = self.parent.ships
            rect.width = int(self.ox) * 2
            distance = abs(ships[0].location.centery - \
                           ships[1].location.centery)
            rect.height = distance + int(self.oy) * 2
            rect.centerx = ships[0].location.centerx
            if ships[0].location.centery > ships[1].location.centery:
                rect.centery = ships[1].location.centery + (distance / 2)
            else:
                rect.centery = ships[0].location.centery + (distance / 2)
            color = Color(0, 0, 0)
            color.hsla = self.hues[0], 90, 70, 100
            polygon(self.display_surface, color,
                    (rect.topleft, rect.topright, rect.bottomright,
                     rect.bottomleft), 1)
            self.hues.rotate(-1)
            color.hsla = self.hues[0], 90, 90, 100
            inner = rect.inflate(-4, -4)
            polygon(self.display_surface, color,
                    (inner.topleft, inner.topright, inner.bottomright,
                     inner.bottomleft), 1)


class Enemies(GameChild, list):

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        self.hit_sound_effect = SoundEffect(self.get_resource("hit.wav"), 1)

    def reset(self):
        list.__init__(self, [])
        self.halt_spawning()

    def halt_spawning(self):
        self.spawning = False

    def spawn_initial(self):
        for ii in xrange(self.parent.INITIAL_POPULATION):
            self.append(Enemy(self))

    def activate_spawning(self):
        self.spawning = True

    def update(self):
        if self.spawning and random() < self.parent.ENEMY_SPAWN_RATE:
            self.append(Enemy(self))
        if not self.parent.game_over:
            struck = False
            if self.parent.net.active:
                for enemy in self:
                    for location in enemy.locations:
                        if location.colliderect(self.parent.net.rect):
                            self.remove(enemy)
                            if not struck:
                                struck = True
                                self.parent.life_meter.remove()
                                self.hit_sound_effect.play()
                if struck:
                    self.parent.net.deactivate()
            if not struck:
                for enemy in self:
                    for location in enemy.locations:
                        if location.collidelist([ship.location for ship in \
                                                 self.parent.ships]) != -1:
                            self.remove(enemy)
                            if not struck:
                                struck = True
                                self.parent.life_meter.remove()
                                self.hit_sound_effect.play()
        for enemy in self:
            enemy.update()


class Enemy(Sprite):

    def __init__(self, parent):
        Sprite.__init__(self, parent)
        self.display_surface = self.get_display_surface()
        self.speed = self.parent.parent.ENEMY_SPEED
        self.set_frames()
        self.place()

    def set_frames(self):
        w = 21
        color = Color(0, 0, 0)
        for hue in 0, 10, 20, 30, 40, 50, 60, 50, 40, 30, 20, 10:
            frame = Surface((w, w))
            frame.set_colorkey((0, 0, 0))
            color.hsla = hue, 100, 50, 100
            polygon(frame, color, ((0, w - 1), (w / 2, 0), (w - 1, w - 1)))
            for x in xrange(1, w, 2):
                frame.fill((0, 0, 0), (x, 0, 1, w))
            self.add_frame(frame)

    def place(self):
        self.location.center = self.get_random_location()
        while self.location.collidelist([ship.location for ship in \
                                         self.parent.parent.ships]) != -1:
            self.location.center = self.get_random_location()
        self.set_course()
        ds = self.display_surface
        self.add_location(offset=(ds.get_width(), 0))
        self.add_location(offset=(0, ds.get_height()))

    def get_random_location(self):
        return randint(0, self.display_surface.get_width()), \
               randint(0, self.display_surface.get_height())

    def set_course(self):
        self.course = [randint(21, 395), radians(randrange(0, 360))]

    def update(self):
        if self.course:
            if self.course[0] < self.speed:
                self.course = None
            else:
                self.move(sin(self.course[1]) * self.speed,
                          cos(-self.course[1]) * self.speed)
                self.course[0] -= self.speed
            rect = self.display_surface.get_rect()
            if self.location.right < 0:
                self.move(rect.w)
            if self.location.left > rect.w:
                self.move(-rect.w)
            if self.location.bottom < 0:
                self.move(dy=rect.h)
            if self.location.top > rect.h:
                self.move(dy=-rect.h)
        if self.course is None and random() < \
               self.parent.parent.ENEMY_MOVE_PROBABILITY:
            self.set_course()
        Sprite.update(self)


class LifeMeter(GameChild):

    MAX_LIVES = 3

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        self.display_surface = self.get_display_surface()
        w, h = 75, 25
        background = self.background = Surface((w, h))
        background.fill((255, 255, 255))
        background.fill((128, 90, 35), (0, h - 2, w, 2))
        rect = self.background_rect = background.get_rect()
        rect.midbottom = self.display_surface.get_rect().midbottom
        self.heart = load(self.get_resource("heart.png")).convert_alpha()
        self.blank = load(self.get_resource("blank.png")).convert_alpha()
        self.sprite_rect = self.heart.get_rect()
        self.sprite_rect.centery = rect.centery

    def reset(self):
        self.lives = self.MAX_LIVES

    def remove(self):
        self.lives -= 1

    def update(self):
        self.display_surface.blit(self.background, self.background_rect)
        for ii in range(self.MAX_LIVES):
            rect = self.sprite_rect
            rect.centerx = self.background_rect.centerx + \
                               (ii - ((self.MAX_LIVES - 1) / 2)) * \
                               self.background_rect.w / 3
            if ii < self.lives:
                self.display_surface.blit(self.heart, rect)
            else:
                self.display_surface.blit(self.blank, rect)


class Grid(GameChild, list):

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        self.display_surface = self.get_display_surface()
        background = self.background = Surface((self.parent.life_meter.\
                                                background_rect.w, 14))
        rect = self.background_rect = background.get_rect()
        rect.bottomleft = self.parent.life_meter.background_rect.topleft
        self.font = Font(self.get_resource("display", "completion-font"), 12)
        self.font.set_bold(True)

    def reset(self):
        ds = self.display_surface
        list.__init__(self, [[False for y in xrange(ds.get_height())] for x in \
                             xrange(ds.get_width())])
        self.ratio = 0.0
        self.paint()

    def paint(self):
        color = Color(0, 0, 0)
        color.hsla = int(120 * self.ratio), 100, 50, 100
        self.background.fill(color)

    def add(self, rect):
        for x in xrange(rect.left, rect.right):
            self[x][rect.top:rect.bottom] = [True] * (rect.bottom - rect.top)
        count = 0
        for x in self.parent.grid:
            for y in x:
                count += y
        self.ratio = float(count) / (len(self) * len(self[0]))
        self.paint()

    def update(self):
        self.display_surface.blit(self.background, self.background_rect)
        w = 10
        display = str(int(self.ratio * 100)) + "%"
        for ii, char in enumerate(display):
            glyph = self.font.render(char, True, (255, 255, 255))
            rect = glyph.get_rect()
            rect.centery = self.background_rect.centery
            rect.centerx = self.background_rect.centerx + \
                           (ii - ((len(display) - 1) / 2.0)) * w
            self.display_surface.blit(glyph, rect)


class PowerUps(GameChild, list):

    POSITIVE_COLORS = Color(0, 0, 0), Color(255, 255, 255), \
                      (Color(255, 255, 0), Color(0, 255, 0), Color(0, 255, 255))
    NEGATIVE_COLORS = Color(255, 255, 255), Color(0, 0, 0), \
                      (Color(255, 0, 0), Color(153, 0, 153), Color(255, 128, 0))

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        for path in glob(join(self.get_resource("power-up"), "*.png")):
            self.append(PowerUp(self, load(path).convert_alpha(),
                                basename(path).replace("-", " ").split(".")[0]))
        self.history = deque(maxlen=3)

    def reset(self):
        self.deactivate()

    def deactivate(self):
        self.active = False

    def activate(self):
        self.active = True
        for power_up in self:
            power_up.deactivate()

    def update(self):
        if self.active:
            if len(self) < 100 and random() < self.parent.POWER_UP_SUPPLY_RATE:
                incoming = choice(self)
                # while incoming in self.history:
                #     incoming = choice(self)
                incoming.generate()
                self.history.append(incoming)
            for power_up in self:
                power_up.update()


class PowerUp(Sprite):

    BLINK_LENGTH = 1500
    AVAILABILITY = 8900
    SPEED = 1

    def __init__(self, parent, foreground, power):
        Sprite.__init__(self, parent, 240)
        self.foreground = foreground
        self.power = power
        self.display_surface = self.get_display_surface()
        self.time_filter = self.get_game().time_filter
        self.set_frames()
        self.deactivate()

    def set_frames(self):
        foreground_color, background_color, border_colors = \
                          self.parent.NEGATIVE_COLORS if self.power in \
                          ["spawn enemy", "speed up"] else self.parent.\
                          POSITIVE_COLORS
        foreground = self.foreground
        pixels = PixelArray(foreground)
        pixels.replace((0, 0, 0), foreground_color)
        del pixels
        rect = foreground.get_rect().inflate(2, 2)
        rect.topleft = 0, 0
        background = Surface(rect.size)
        background.fill(background_color)
        print rect
        for color in border_colors:
            frame = background.copy()
            frame.blit(foreground, (1, 1))
            frame.fill(color, (0, 0, rect.w, 1))
            frame.fill(color, (0, rect.w - 1, 1, rect.h))
            frame.fill(color, (rect.h - 1, 0, rect.w, 1))
            frame.fill(color, (0, 0, 1, rect.h))
            self.add_frame(frame)

    def deactivate(self):
        self.active = False
        self.remove_locations()

    def generate(self):
        ds = self.display_surface.get_rect()
        coordinates = randrange(0, ds.w), randrange(0, ds.h)
        if not self.active:
            self.active = True
            self.location.topleft = coordinates
        else:
            self.add_location(coordinates)
        location = self.locations[-1]
        location.blinking = True
        location.blink_length = 0
        location.display_length = 0
        location.angle = radians(randrange(0, 360))
        location.curve = uniform(-.05, .05)

    def update(self):
        if self.active:
            frame_duration = self.time_filter.get_last_frame_duration()
            width, height = self.display_surface.get_size()
            for location in self.locations:
                if location.blinking:
                    location.blink_length += frame_duration
                    if location.blink_length >= self.BLINK_LENGTH:
                        location.blinking = False
                        location.unhide()
                    else:
                        location.toggle_hide_state()
                elif location.display_length >= self.AVAILABILITY:
                    if len(self.locations) > 1:
                        self.remove_locations(location)
                    else:
                        self.deactivate()
                else:
                    location.display_length += frame_duration
                    location.move_ip(sin(location.angle) * self.SPEED,
                                     cos(-location.angle) * self.SPEED)
                    location.angle += triangular(-.03 + location.curve,
                                                 .03 + location.curve)
                    if location.right < 0:
                        location.move_ip(width, 0)
                    if location.left > width:
                        location.move_ip(width, 0)
                    if location.bottom < 0:
                        location.move(0, height)
                    if location.top > height:
                        location.move(0, height)
            location_count = len(self.locations)
            for ii in xrange(location_count):
                if not self.locations[ii].is_hidden():
                    self.add_location(offset=(width, 0), base=ii)
                    self.add_location(offset=(0, height), base=ii)
            Sprite.update(self)
            self.locations = self.locations[:location_count]


class EndMessage(Sprite):

    def __init__(self, parent, text, hue):
        Sprite.__init__(self, parent)
        self.font = Font(self.get_resource("display", "end-message-font"), 32)
        spaced = text[0]
        for char in text[1:]:
            spaced += "   " + char
        self.set_frames(spaced, hue)
        self.location.center = self.get_display_surface().get_rect().center

    def reset(self):
        self.deactivate()

    def deactivate(self):
        self.active = False

    def activate(self):
        self.active = True

    def set_frames(self, text, hue):
        color = Color(0, 0, 0)
        for lightness in xrange(50, 100, 5):
            color.hsla = hue, 100, lightness, 100
            self.add_frame(self.font.render(text, True, color))

    def update(self):
        if self.active:
            Sprite.update(self)


class Instructions(GameChild):

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        ds = self.display_surface = self.get_display_surface()
        image = load(self.get_resource("instructions.png")).convert_alpha()
        foreground = self.foreground = Surface(image.get_size(), SRCALPHA)
        fr = self.foreground_rect = foreground.get_rect()
        fr.center = ds.get_rect().center
        color = Color(0, 0, 0)
        # color.hsla = 350, 60, 92, 100
        color.hsva = 87, 57, 89, 100
        foreground.fill(color)
        foreground.blit(image, (0, 0))
        shadow = self.shadow = Surface(foreground.get_size())
        # color.hsla = 350, 60, 26, 100
        color.hsva = 345, 57, 25, 100
        shadow.fill(color)
        sr = self.shadow_rect = shadow.get_rect()
        sr.center = fr.centerx - 8, fr.centery + 8

    def reset(self):
        self.activate()

    def activate(self):
        self.active = True

    def deactivate(self):
        self.active = False

    def update(self):
        if self.active:
            self.display_surface.blit(self.shadow, self.shadow_rect)
            self.display_surface.blit(self.foreground, self.foreground_rect)


class HighScores(Animation):

    def __init__(self, parent):
        Animation.__init__(self, parent)
        self.display_surface = self.get_display_surface()
        self.register(self.blink, interval=500)
        self.refresh()
        self.play(self.blink)

    def reset(self):
        self.activate()
        self.blink_on = True

    def activate(self):
        self.active = True

    def deactivate(self):
        self.active = False

    def refresh(self, submitted=None):
        scores = map(float, sorted(open(self.get_resource("scores")).\
                                   readlines(), reverse=True)[:5])
        path = self.get_resource("display", "high-scores-font")
        font = Font(path, 14)
        font.set_italic(True)
        font.set_bold(True)
        foreground_color = (192, 127, 40)
        background_color = (254, 220, 178)
        plates = self.plates = [font.render(" BEST", True, foreground_color,
                                            background_color)]
        self.cancel_blink()
        for ii, score in enumerate(scores):
            size = range(22, 12, -2)[ii]
            if str(score) == submitted:
                self.blink_index = ii
            text = "%.2f" % (score * 100)
            font = Font(path, size)
            font.set_italic(True)
            main = font.render(text[:2], True, foreground_color)
            precision = Font(path, size - 6).render(text[-3:], True,
                                                    foreground_color)
            background = Surface((main.get_width() + precision.get_width() + 1,
                                  main.get_height()))
            background.fill(background_color)
            background.blit(main, (2, 1))
            rect = precision.get_rect()
            br = background.get_rect()
            rect.bottomright = br.right - 2, br.bottom
            background.blit(precision, rect)
            plates.append(background)

    def cancel_blink(self):
        self.blink_index = None

    def blink(self):
        self.blink_on = not self.blink_on

    def update(self):
        if self.active:
            step = self.display_surface.get_height() / (len(self.plates) + 1)
            for ii, y in enumerate(xrange(step,
                                          self.display_surface.get_height() - \
                                          step, step)):
                plate = self.plates[ii]
                rect = plate.get_rect()
                rect.midleft = 0, y
                if self.blink_on or ii - 1 != self.blink_index:
                    self.display_surface.blit(plate, rect)
            Animation.update(self)
